
from __future__ import annotations

from contextlib import AsyncExitStack
from typing import TypeVar

from pymap.context import connection_exit

__all__ = ['exit_context']

_F = TypeVar('_F')


def exit_context(func: _F) -> _F:
    async def deco(*args, **kwargs):
        async with AsyncExitStack() as stack:
            connection_exit.set(stack)
            return await func(*args, **kwargs)
    return deco  # type: ignore




from __future__ import annotations

import asyncio
import os
import os.path
import tempfile
from argparse import ArgumentParser, Namespace
from asyncio import Task, AbstractServer, CancelledError
from typing import Optional

from grpclib.server import Server
from pymap.config import IMAPConfig
from pymap.interfaces.backend import BackendInterface, ServiceInterface

from .handlers import AdminHandlers

__all__ = ['AdminService']


class AdminService(ServiceInterface):  # pragma: no cover
    """A pymap service implemented using a `grpc <https://grpc.io/>`_ server
    to perform admin functions on a running backend.

    """

    def __init__(self, path: str, server: AbstractServer) -> None:
        super().__init__()
        self._path = path
        self._server = server
        self._task = asyncio.create_task(self._run())

    @classmethod
    def get_socket_path(cls) -> str:
        """Return the default location of the UNIX socket file listening for
        admin requests.

        """
        return os.path.join(tempfile.gettempdir(), 'pymap-admin.sock')

    @classmethod
    def add_arguments(cls, parser: ArgumentParser) -> None:
        group = parser.add_argument_group('admin service')
        group.add_argument('--admin-socket', metavar='PATH', dest='admin_sock',
                           help='path to socket file')
        group.add_argument('--no-filter', action='store_true',
                           help='do not filter appended messages')

    @classmethod
    async def start(cls, backend: BackendInterface,
                    config: IMAPConfig) -> AdminService:
        config = backend.config
        path: Optional[str] = config.args.admin_sock
        handlers = AdminHandlers(backend)
        server = Server([handlers], loop=asyncio.get_event_loop())
        path = await cls._start(path, server)
        cls._chown(path, config.args)
        return cls(path, server)

    @classmethod
    async def _start(cls, path: Optional[str], server: Server) -> str:
        if not path:
            path = cls.get_socket_path()
        await server.start(path=path)
        return path

    @classmethod
    def _chown(cls, path: str, args: Namespace) -> None:
        uid = args.set_uid or -1
        gid = args.set_gid or -1
        if uid >= 0 or gid >= 0:
            os.chown(path, uid, gid)

    @property
    def task(self) -> Task:
        return self._task

    def _unlink_path(self) -> None:
        try:
            os.unlink(self._path)
        except OSError:
            pass

    async def _run(self) -> None:
        server = self._server
        try:
            await server.wait_closed()
        except CancelledError:
            server.close()
            await server.wait_closed()
        finally:
            self._unlink_path()



"""Admin functions for a running pymap server."""

from __future__ import annotations

import asyncio
import sys
from argparse import ArgumentParser, Namespace, FileType
from typing import Type, Mapping

from grpclib.client import Channel
from pkg_resources import iter_entry_points, DistributionNotFound
from pymap import __version__

from .command import ClientCommand
from .. import AdminService
from ..grpc.admin_grpc import AdminStub


def main() -> int:
    parser = ArgumentParser(description=__doc__)
    parser.add_argument('--version', action='version',
                        version='%(prog)s' + __version__)
    parser.add_argument('--outfile', metavar='PATH',
                        type=FileType('w'), default=sys.stdout,
                        help='the output file (default: stdout)')
    parser.add_argument('--socket', metavar='PATH', help='path to socket file')

    subparsers = parser.add_subparsers(dest='command',
                                       help='which admin command to run')
    commands = _load_entry_points('pymap.admin.client')
    for command_name, command_cls in commands.items():
        command_cls.add_subparser(command_name, subparsers)
    args = parser.parse_args()

    if not args.command:
        parser.error('Expected command name.')
    command = commands[args.command]

    return asyncio.run(run(parser, args, command), debug=False)


async def run(parser: ArgumentParser, args: Namespace,
              command_cls: Type[ClientCommand]) -> int:
    loop = asyncio.get_event_loop()
    path = args.socket or AdminService.get_socket_path()
    channel = Channel(path=path, loop=loop)
    stub = AdminStub(channel)
    command = command_cls(stub, args)
    try:
        code = await command.run(args.outfile)
    finally:
        channel.close()
    return code


def _load_entry_points(group: str) -> Mapping[str, Type[ClientCommand]]:
    ret = {}
    for entry_point in iter_entry_points(group):
        try:
            cls = entry_point.load()
        except DistributionNotFound:
            pass  # optional dependencies not installed
        else:
            ret[entry_point.name] = cls
    return ret
